import os
import uuid
import yaml
import logging
import re
from functools import reduce
from datetime import datetime
from typing import List, Optional
from box import Box
from gluepy.utils.loading import LazyProxy, import_string, SingletonMixin
from gluepy.utils.dict import merge
from gluepy.conf import default_settings

logger = logging.getLogger(__name__)


class Context(SingletonMixin):
    _data = Box()

    def __init__(self, data=None, *args, **kwargs):
        if data:
            self._data.clear()
            self._data.update(data)

    @property
    def instance(self):
        return self.__singleton_instance

    def __str__(self):
        return str(self._data)

    def __hash__(self):
        return hash(self._data)

    def __dir__(self):
        return dir(self._data)

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __setattr__(self, name, value):
        raise TypeError(
            f"'{self.__class__.__name__}' object does not support item assignment"
        )


class DefaultContextManager:
    """
    Class responsible for populating the ``Context`` object that is
    later used by ``default_context``.

    The default context manager is populating the context object by reading
    in a set of YAML files located in the :setting:`CONFIG_PATH` directory.


    """

    def __init__(self) -> None:
        self._ctx = None

    def create_context(
        self,
        run_id: Optional[str] = None,
        run_folder: Optional[str] = None,
        patches: Optional[List[str]] = None,
        evaluate_lazy: bool = False,
    ):
        """Create a new context instance.

        Read in existing configuration files from :setting:`CONFIG_PATH` directory
        and patch them with any overrides from ``patches`` kwarg passed in.

        Also populates the ``meta`` key of the context with run meta data
        such as run_folder, run_id and created_at timestamp.

        """
        # Imported here to avoid circular import.
        from gluepy.files.storages import default_storage

        patches = patches or []
        if patches:
            for i, path in enumerate(patches):
                if not default_storage.exists(path):
                    logger.warning(f"Patch '{path}' was not found.")
                    patches[i] = dict()
                    continue

                patches[i] = yaml.load(
                    default_storage.open(path),
                    Loader=yaml.SafeLoader,
                )

        yamls = filter(
            lambda f: os.path.splitext(f)[1] in {".yaml", ".yml"},
            os.listdir(default_settings.CONFIG_PATH),
        )

        patches = [
            yaml.load(
                open(os.path.join(default_settings.CONFIG_PATH, y)),
                Loader=yaml.SafeLoader,
            )
            for y in yamls
        ] + patches

        patches += [self.get_run_meta(run_id=run_id, run_folder=run_folder)]

        ctx = Context(
            data=reduce(lambda config, patch: merge(config, patch), patches, {})
        )

        if evaluate_lazy:
            self._ctx = ctx
        return ctx

    def load_context(self, path: str, patches: Optional[List[str]] = None):
        """Loads an existing context.

        When we may want to recreate or rerun a pre-existing model execution,
        we may want to load the same context and parameters used by that pre-existing
        execution.

        This method reads in an existing file and applies a set of optional patches
        as overrides, and will reuse the same run_id, run_folder as is already
        defined in the pre-existing context YAML file passed in.

        """
        # Imported here to avoid circular import.
        from gluepy.files.storages import default_storage

        patches = patches or []
        # Reformat folder name to be valid.
        path_formatted = re.sub(r"[:\+\s]+", "-", os.path.dirname(path))
        path_formatted = re.sub(r"[^a-zA-Z0-9-_/\.]+", "", path_formatted)
        # Reconstrut the newly formatted dir path with the file name.
        path_formatted = os.path.join(path_formatted, os.path.basename(path))
        if path != path_formatted:
            logger.warning(
                f"Given run folder ('{path}') path was "
                f"reformatted into '{path_formatted}'"
            )

        if not default_storage.exists(path_formatted):
            self._ctx = self.create_context(
                run_id=(
                    os.path.dirname(path_formatted.rstrip(os.sep)).split(os.sep)[-1]
                    or str(uuid.uuid4())
                ),
                run_folder=os.path.dirname(path_formatted),
                patches=patches,
            )
            return self._ctx

        _, ext = os.path.splitext(path_formatted)
        if ext not in {".yaml", ".yml"}:
            raise ValueError(f"File path '{path_formatted}' is not a YAML file.")

        self._ctx = Context(
            yaml.load(default_storage.open(path_formatted), Loader=yaml.SafeLoader)
        )
        return self._ctx

    def get_run_meta(self, **kwargs):
        dt = datetime.utcnow()
        run_id = kwargs.get("run_id") or str(uuid.uuid4())
        run_folder = kwargs.get("run_folder") or os.path.join(
            f"runs/{dt.year}/{dt.month}/{dt.day}", run_id
        )
        return {
            "gluepy": {
                "run_id": run_id,
                "created_at": dt,
                "run_folder": run_folder,
            }
        }


default_context_manager = LazyProxy(
    lambda: import_string(default_settings.CONTEXT_BACKEND)()
)
default_context = LazyProxy(
    lambda: default_context_manager._ctx
    if default_context_manager._ctx
    else default_context_manager.create_context()
)



import os
import logging
from typing import List, Optional
import time
import click
from gluepy.conf import default_context_manager
from gluepy.files.storages import default_storage
from . import cli

logger = logging.getLogger(__name__)


@cli.command()
@click.option("--task", type=str)
@click.option("--from-task", type=str)
@click.option("--patch", "-p", type=str, multiple=True)
@click.option("--retry", type=str)
@click.argument("label")
def dag(
    label,
    retry: Optional[str] = None,
    patch: Optional[List[str]] = None,
    from_task: Optional[str] = None,
    task: Optional[str] = None,
):
    """Command to run a DAG by its label.

    Args:
        label (str): The label of the DAG to execute.
        retry (Optional[str], optional): Path to existing run_folder of a run to retry.
          Defaults to None.
        patch (Optional[List[str]], optional): Path to patch YAML file to override
            context with. Defaults to None.
        from_task (Optional[str], optional): Label of task in DAG to retry from.
            Defaults to None.
        task (Optional[str], optional): Label of task if only want to execute a
            single task in DAG. Defaults to None.

    """
    DAG = _get_dag_by_label(label)
    assert not (from_task and task), "Only one of --from-task or --task can be set."
    retry = retry if retry is None else retry.strip(default_storage.separator)

    if retry and default_storage.exists(os.path.join(retry, "context.yaml")):
        default_context_manager.load_context(
            os.path.join(retry, "context.yaml"), patches=list(patch)
        )
    elif retry:
        # Retry a run folder path that does not exist by creating it.
        default_context_manager.create_context(
            run_id=os.path.basename(retry),
            run_folder=retry,
            patches=list(patch) if patch else None,
            evaluate_lazy=True,
        )
    elif patch:
        default_context_manager.create_context(patches=list(patch), evaluate_lazy=True)

    tasks = DAG().inject_tasks()

    if task:
        tasks = [_get_task_by_label(task)]

    if from_task:
        Task = _get_task_by_label(from_task)
        for i, t in enumerate(tasks):
            if t is Task:
                pos = i
                break
        else:
            raise ValueError(f"Task '{from_task}' not found in DAG list of tasks.")
        tasks = tasks[pos:]

    for t in tasks:
        logger.info(f"---------- Started task '{t.__name__}'")
        time_start = time.time()
        t().run()
        time_end = time.time()
        logger.info(
            f"---------- Completed task '{t.__name__}' in "
            f"{'{:f}'.format(time_end-time_start)} seconds"
        )


def _get_dag_by_label(label):
    from gluepy.exec import DAG_REGISTRY

    try:
        DAG = DAG_REGISTRY[label]
    except KeyError:
        raise ValueError(f"DAG with label '{label}' was not found in registry.")

    return DAG


def _get_task_by_label(label):
    from gluepy.exec import TASK_REGISTRY

    try:
        Task = TASK_REGISTRY[label]
    except KeyError:
        raise ValueError(f"Task with label '{label}' was not found in registry.")

    return Task



import os
from typing import List, Tuple, Union
from pathlib import Path
from io import StringIO, BytesIO
from gluepy.conf import default_settings, default_context


class BaseStorage:
    """Base class of a Storage implementation"""

    MAX_CHUNK_SIZE = 1_000_000
    separator = os.sep

    def abspath(self, path: str) -> str:
        """Get absolute path to file including STORAGE_ROOT"""
        return str(Path(default_settings.STORAGE_ROOT) / path)

    def relpath(self, path: str) -> str:
        """Get relative path to file, relative to STORAGE_ROOT"""
        return str(Path(path).relative_to(default_settings.STORAGE_ROOT))

    def runpath(self, path: str) -> str:
        """Get the path appended to the current context's :ref:`context_run_folder`"""
        return str(Path(default_context.gluepy.run_folder) / path)

    def _to_bytes(self, content: StringIO):
        """Write a StringIO stream to BytesIO"""
        io = BytesIO()
        while True:
            chunk = content.read(self.MAX_CHUNK_SIZE)
            if not chunk:
                break

            io.write(bytes(chunk.encode("utf-8")))

        io.seek(0, os.SEEK_SET)
        return io

    def touch(self, file_path: str, content: Union[StringIO, BytesIO]) -> None:
        """Create a new blob at file path.

        Args:
            file_path (str): Path to file we want to create
            content (Union[StringIO, BytesIO]): Content of file we want to generate
        """
        raise NotImplementedError()

    def open(self, file_path: str, mode: str = "rb") -> Union[str, bytes]:
        """Opens a blob at file_path

        Args:
            file_path (str): File path of blob we want to open

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        raise NotImplementedError()

    def rm(self, path: str, recursive: bool = False) -> None:
        """Delete a file

        Args:
            path (str): Path to file to delete
            recursive (bool): If allowed to delete recursive directories or not.
        """
        raise NotImplementedError()

    def cp(
        self,
        src_path: str,
        dest_path: str,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Copy a file from source to destination

        Args:
            src_path (str): Path to file or directory to copy.
            dest_path (str): Path to file or directory to copy to.
            recursive (bool): If should copy sub directories as well.
            overwrite (bool): If should copy to destination that already exists.
        """
        raise NotImplementedError()

    def ls(self, path: str) -> Tuple[List[str], List[str]]:
        """List all files and directories at given path.

        Args:
            path (str): Path where we want to list contents of

        Returns:
            Tuple[List[str], List[str]]:
              First list is files, second list is directories.
        """
        raise NotImplementedError()

    def mkdir(self, path: str, make_parents: bool = False) -> None:
        """Make a new directory at location

        Args:
            path (str): Path of directory we want to create
            make_parents (bool, optional): If we should generate parents
            folders as well. Defaults to False.
        """
        raise NotImplementedError()

    def isdir(self, path: str) -> bool:
        """Check if path is directory or not.

        Args:
            path (str): Path we want to check

        Returns:
            bool: True/False if path is directory or not
        """
        raise NotImplementedError()

    def isfile(self, path: str) -> bool:
        """Check if path is a file or not.

        Args:
            path (str): Path we want to check

        Returns:
            bool: True/False if path is file or not.
        """
        raise NotImplementedError()

    def exists(self, path: str) -> bool:
        """Check if path exists or not.

        Args:
            path (str): Path we want to check

        Returns:
            bool: True/False if path is file or not.
        """
        raise NotImplementedError()



import logging
from io import BytesIO
import os
import pandas as pd
from gluepy.files.data import BaseDataManager
from gluepy.conf import default_settings
from gluepy.files.storages import default_storage

logger = logging.getLogger(__name__)


class PandasDataManager(BaseDataManager):
    """Data Manager that implement read and write actions for
    pandas dataframes on the currently set storage backend.
    """

    def read(self, path: str, root: bool = False, *args, **kwargs) -> pd.DataFrame:
        """Read in a pandas dataframe from path.

        Args:
            path (str): Path to file to read in.
            root (bool, optional): If path is relative to root or run folder.
                Defaults to False.

        Raises:
            ValueError: Raised if file extension is not supported.

        Returns:
            pd.DataFrame: Loaded pandas dataframe
        """
        _, ext = os.path.splitext(path)
        if ext in {".csv", ".txt"}:
            return self._read_csv(path, root, *args, **kwargs)
        elif ext in {".pq", ".parquet"}:
            return self._read_parquet(path, root, *args, **kwargs)
        elif ext in {
            ".json",
        }:
            return self._read_json(path, root, *args, **kwargs)
        else:
            raise ValueError(
                f"'{self.__class__.__name__}' does not support "
                f"reading files of extension '{ext}'"
            )

    def read_sql(self, sql: str, *args, **kwargs) -> pd.DataFrame:
        try:
            import pandas_gbq
        except ImportError:
            logger.error("Dependency 'pandas_gbq' must be installed to read SQL")
            raise
        return pandas_gbq.read_gbq(
            sql,
            project_id=default_settings.GCP_PROJECT_ID,
            use_bqstorage_api=True,
            *args,
            **kwargs,
        )

    def write(
        self, path: str, df: pd.DataFrame, root: bool = False, *args, **kwargs
    ) -> None:
        """Write pandas dataframe to path on default storage backend.

        Args:
            path (str): Path to destination of dataframe.
            df (pd.DataFrame): Dataframe to be written.
            root (bool, optional): If path is relative to root or run folder.
                Defaults to False.

        Raises:
            ValueError: Raised if file extension is not supported.
        """
        _, ext = os.path.splitext(path)
        if ext in {".csv", ".txt"}:
            self._write_csv(path, df, root, *args, **kwargs)
        elif ext in {".pq", ".parquet"}:
            self._write_parquet(path, df, root, *args, **kwargs)
        else:
            raise ValueError(
                f"'{self.__class__.__name__}' does not support "
                f"writing files of extension '{ext}'"
            )

    def _read_csv(self, path: str, root: bool = False, *args, **kwargs):
        """Implementation of reading csv file"""
        stream = BytesIO(
            default_storage.open(
                path if root is True else default_storage.runpath(path)
            )
        )
        logger.info(f"Reading file from path '{path}'.")
        return pd.read_csv(stream, *args, **kwargs)

    def _read_parquet(self, path: str, root: bool = False, *args, **kwargs):
        """Implementation of reading parquet file"""
        stream = BytesIO(
            default_storage.open(
                path if root is True else default_storage.runpath(path)
            )
        )
        logger.info(f"Reading file from path '{path}'.")
        return pd.read_parquet(stream, *args, **kwargs)

    def _read_json(self, path: str, root: bool = False, *args, **kwargs):
        """Implementation of reading json file"""
        stream = BytesIO(
            default_storage.open(
                path if root is True else default_storage.runpath(path)
            )
        )
        logger.info(f"Reading file from path '{path}'.")
        return pd.read_json(stream, *args, **kwargs)

    def _write_csv(
        self, path: str, df: pd.DataFrame, root: bool = False, *args, **kwargs
    ):
        """Implementation of writing csv file"""
        path = path if root is True else default_storage.runpath(path)
        logger.info(f"Writing file to path '{path}'.")
        stream = BytesIO()
        df.to_csv(stream, *args, **kwargs)
        stream.seek(0, os.SEEK_SET)

        default_storage.touch(file_path=path, content=stream)

    def _write_parquet(
        self, path: str, df: pd.DataFrame, root: bool = False, *args, **kwargs
    ):
        """Implementation of reading parquet file"""
        path = path if root is True else default_storage.runpath(path)
        logger.info(f"Writing file to path '{path}'.")
        stream = BytesIO()
        df.to_parquet(stream, *args, **kwargs)
        stream.seek(0, os.SEEK_SET)

        default_storage.touch(file_path=path, content=stream)



def merge(a, b, path: str = None, allow_new_key: bool = True):
    """Recursively merge two dictionaries"""
    path = path or []
    a = {} if a is None else a
    b = {} if b is None else b

    for key in b:
        if key in a:
            if (
                a[key] is not None
                and b[key] is not None
                and not isinstance(a[key], type(b[key]))
            ):
                raise TypeError(
                    f"Conflicting types '{a[key].__class__.__name__}' and "
                    f"'{b[key].__class__.__name__}' at '{'.'.join(path + [str(key)])}'"
                )

            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)], allow_new_key)
            else:
                a[key] = b[key]
        elif not allow_new_key:
            raise ValueError(
                f"Key '{'.'.join(path + [str(key)])}' missing in source dictionary"
            )
        else:
            a[key] = b[key]

    return a



import os
from typing import Any
from gluepy.utils.loading import LazyProxy, import_module, SingletonMixin


class Settings(SingletonMixin):
    """Gluepy settings object

    This singleton settings object captures all the current
    active settings of the project and execution. It will load all
    settings using the GLUEPY_SETTINGS_MODULE environment variable,
    which allow you to have multiple different setting files that
    can be loaded by configuring your environment.

    """

    BASE_DIR: str
    CONFIG_PATH: str
    INSTALLED_MODULES: list[str]
    STORAGE_BACKEND: str
    STORAGE_ROOT: str
    DATA_BACKEND: str
    CONTEXT_BACKEND: str
    START_TASK: str
    AIRFLOW_DAG_PREFIX: str
    AIRFLOW_TEMPLATE: str
    AIRFLOW_IMAGE: str
    AIRFLOW_CONFIGMAPS: list[str]
    AIRFLOW_POD_RESOURCES: dict[str, dict]
    AIRFLOW_KUBERNETES_CONFIG: str
    LOGGING: dict[str, Any]

    def __init__(self, dotted_path: str, *args, **kwargs):
        module = import_module(dotted_path)
        for key in dir(module):
            if key.isupper():
                setattr(self, key, getattr(module, key))


default_settings = LazyProxy(lambda: Settings(os.environ.get("GLUEPY_SETTINGS_MODULE")))



import threading
import operator
from importlib import import_module


# User to separate None from not invoked yet (still lazy)
empty = object()


def new_method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    return inner


class LazyProxy(object):

    _wrapped = None

    def __init__(self, factory):
        self.__dict__["_factory"] = factory
        self._wrapped = empty

    def _setup(self):
        self._wrapped = self._factory()

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == "_wrapped":
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == "_wrapped":
            raise ValueError("Cannot delete `_wrapped_`")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __func__ = None
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __gt__ = new_method_proxy(operator.gt)
    __ne__ = new_method_proxy(operator.ne)
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err


class SingletonMixin(object):
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def instance(cls):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = cls()
        return cls.__singleton_instance



import os
import logging
from typing import List, Tuple, Union
from pathlib import Path
from io import StringIO, BytesIO
from gluepy.conf import default_settings
from gluepy.exceptions import BootstrapError

try:
    import boto3
    from botocore.client import Config
    from botocore.exceptions import ClientError
except ImportError as e:
    raise BootstrapError("Could not load Boto3's S3 bindings. %s" % e)
from gluepy.files.storages.base import BaseStorage

logger = logging.getLogger(__name__)


class S3Storage(BaseStorage):
    """
    Storage backend to support S3 based storages such as
    AWS S3 or DigitalOcean Spaces.
    """

    MAX_CHUNK_SIZE = 32_000_000

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = None
        self._bucket = None
        self.region_name = default_settings.AWS_S3_REGION_NAME
        self.use_ssl = True
        self.endpoint_url = default_settings.AWS_S3_ENDPOINT_URL
        self.config = Config(
            s3={"addressing_style": None},
            signature_version=None,
            proxies=None,
        )
        self.verify = True
        self.access_key, self.secret_key = (
            default_settings.AWS_ACCESS_KEY_ID,
            default_settings.AWS_SECRET_ACCESS_KEY,
        )

    @property
    def connection(self):
        if self._connection is None:
            session = boto3.session.Session(
                aws_access_key_id=self.access_key,
                aws_secret_access_key=self.secret_key,
            )
            self._connection = session.resource(
                "s3",
                region_name=self.region_name,
                use_ssl=self.use_ssl,
                endpoint_url=self.endpoint_url,
                config=self.config,
                verify=self.verify,
            )

        return self._connection

    @property
    def bucket(self):
        if self._bucket is None:
            self._bucket = self.connection.Bucket(
                default_settings.AWS_STORAGE_BUCKET_NAME
            )
        return self._bucket

    def touch(self, file_path: str, content: Union[StringIO, BytesIO]) -> None:
        """Create a new blob at file path.

        Args:
            file_path (str): Path to file we want to create
            content (Union[StringIO, BytesIO]): Content of file we want to generate

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        if isinstance(content, StringIO):
            content = self._to_bytes(content)
        obj = self.bucket.Object(file_path)
        content.seek(0, os.SEEK_SET)
        obj.upload_fileobj(content)
        return file_path

    def open(self, file_path: str, mode: str = "rb") -> Union[str, bytes]:
        """Opens a blob at file_path

        Args:
            file_path (str): File path of blob we want to open

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        logger.debug(f"Reading file from path '{file_path}'")
        obj = self.bucket.Object(file_path)
        handler = BytesIO()
        obj.download_fileobj(handler)
        handler.seek(0, os.SEEK_SET)
        return (
            handler.read()
            if "b" in mode
            else handler.read().decode("utf-8", errors="ignore")
        )

    def ls(self, path: str) -> Tuple[List[str], List[str]]:
        """List all files and directories at given path.

        Args:
            path (str): Path where we want to list contents of

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            Tuple[List[str], List[str]]:
              First list is files, second list is directories.
        """
        path = self.abspath(path)
        # The path needs to end with a slash, but if the root is empty, leave it.
        if path and not path.endswith("/"):
            path += "/"

        directories = []
        files = []
        paginator = self.connection.meta.client.get_paginator("list_objects")
        pages = paginator.paginate(
            Bucket=default_settings.AWS_STORAGE_BUCKET_NAME,
            Delimiter="/",
            Prefix=path,
            PaginationConfig={"PageSize": 1000},
        )
        for page in pages:
            for entry in page.get("CommonPrefixes", ()):
                directories.append(self.relpath(entry["Prefix"]))
            for entry in page.get("Contents", ()):
                key = entry["Key"]
                if key != path:
                    files.append(self.relpath(key))
        return directories, files

    def mkdir(self, path: str, make_parents: bool = False) -> None:
        """Make a new directory at location

        Args:
            path (str): Path of directory we want to create
            make_parents (bool, optional): If we should generate parents
              folders as well. Defaults to False.

        Raises:
            NotImplementedError: Base class raise NotImplementedError
        """
        self.touch(str(Path(path) / ".empty"))

    def rm(self, path: str, recursive: bool = False) -> None:
        """Delete a file

        Args:
            path (str): Path to file to delete
            recursive (bool): If allowed to delete recursive directories or not.

        """
        if recursive:
            raise NotImplementedError("`recursive` not supported")
        self.bucket.Object(path).delete()

    def cp(
        self,
        src_path: str,
        dest_path: str,
        recursive: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Copy a file from source to destination

        Args:
            src_path (str): Path to file or directory to copy.
            dest_path (str): Path to file or directory to copy to.
            recursive (bool): If should copy sub directories as well.
            overwrite (bool): If should copy to destination that already exists.
        """
        if recursive or overwrite:
            raise NotImplementedError("`recursive` and `overwrite` not supported.")

        self.touch(dest_path, self.open(src_path))

    def isdir(self, path: str) -> bool:
        """Check if path is directory or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is directory or not
        """
        raise NotImplementedError()

    def isfile(self, path: str) -> bool:
        """Check if path is a file or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        raise NotImplementedError()

    def exists(self, path: str) -> bool:
        """Check if path exists or not.

        Args:
            path (str): Path we want to check

        Raises:
            NotImplementedError: Base class raise NotImplementedError

        Returns:
            bool: True/False if path is file or not.
        """
        try:
            self.connection.meta.client.head_object(
                Bucket=default_settings.AWS_STORAGE_BUCKET_NAME, Key=path
            )
            return True
        except ClientError:
            return False



import os
import re
from typing import Optional
from jinja2 import Environment, PackageLoader, FileSystemLoader
from gluepy.conf.context import default_settings
import click
from . import cli


@cli.group()
def airflow():
    """Group for all airflow commands"""


@airflow.command()
@click.argument("dest_path", type=str, required=True)
@click.argument("dag", type=str, required=False)
def generate(dest_path: str, dag: Optional[str] = None):
    """Command used to generate Airflow DAG files from Gluepy DAGs

    Args:
        dest_path (str): Path to store Airflow DAG files generated.
        dag (Optional[str], optional): Label of DAG to generate Airflow DAG of.
            Defaults to None.

    """
    from gluepy.exec.dags import REGISTRY

    if dag and dag not in REGISTRY:
        raise KeyError(f"DAG '{dag}' was not found in registry.")

    dags = [dag] if dag else list(REGISTRY.keys())

    for dag_label in dags:
        dag = REGISTRY[dag_label]()
        label = f"{default_settings.AIRFLOW_DAG_PREFIX}_{dag.label or dag.__class__.__name__.lower()}"  # noqa
        _, template = get_jinja(dag)
        with open(os.path.join(dest_path, f"{label}.py"), mode="w") as handle:
            handle.write(
                template.render(
                    **{
                        "dag": dag,
                        "dag_label": label,
                        "links": dag.inject_tasks(),
                        "image": default_settings.AIRFLOW_IMAGE,
                        "configmaps": default_settings.AIRFLOW_CONFIGMAPS or [],
                        "pod_resources": default_settings.AIRFLOW_POD_RESOURCES or {},
                        "k8s_config": default_settings.AIRFLOW_KUBERNETES_CONFIG
                        or None,
                    }
                )
            )


def get_jinja(dag):
    if dag.extra_options.get("airflow_template"):
        env = Environment(
            loader=FileSystemLoader(default_settings.BASE_DIR), autoescape=True
        )
        env.filters["to_identifier"] = to_identifier
        template = env.get_template(dag.extra_options.get("airflow_template"))
    elif default_settings.AIRFLOW_TEMPLATE:
        env = Environment(
            loader=FileSystemLoader(default_settings.BASE_DIR), autoescape=True
        )
        env.filters["to_identifier"] = to_identifier
        template = env.get_template(default_settings.AIRFLOW_TEMPLATE)
    else:
        env = Environment(
            loader=PackageLoader("gluepy", "templates/airflow"), autoescape=True
        )
        env.filters["to_identifier"] = to_identifier
        template = env.get_template("dag.j2")

    return env, template


def to_identifier(value):
    if not value:
        return None

    value = re.sub(r"[^a-zA-Z0-9_-]", "", value)
    value = re.sub(r"[\s-]", "_", value)

    if not value.isidentifier():
        raise ValueError(f"'{value}' is not a valid python identifier.")

    return value



from typing import Any


class BaseDataManager:
    """Base abstract class for data managers
    that define interface.
    """

    def read(self, path: str, root: bool = False, *args, **kwargs) -> Any:
        """Read dataframe from a file.

        Args:
            path (str): Path to file located on :setting:`STORAGE_BACKEND`.
            root (bool, optional): Is path relative to root or run folder.
                Defaults to False.

        Returns:
            Any: Dataframe depending on implementation.

        """
        raise NotImplementedError()

    def read_sql(self, sql: str, *args, **kwargs) -> Any:
        """Read dataframe from SQL query

        Args:
            sql (str): SQL query to execute to get dataframe.

        Returns:
            Any: Dataframe depending on implementation
        """
        raise NotImplementedError()

    def write(self, path: str, df: Any, root: bool = False, *args, **kwargs) -> None:
        """Write dataframe to file.

        Args:
            path (str): Path to where to write dataframe
            df (Any): Dataframe to be written
            root (bool, optional): If path relatove to root or run folder.
                Defaults to False.

        """
        raise NotImplementedError()



import logging.config
from gluepy.conf import default_settings
from gluepy.utils.loading import import_module


def bootstrap():
    if not hasattr(default_settings, "INSTALLED_MODULES"):
        return

    logging.config.dictConfig(default_settings.LOGGING)
    for module in default_settings.INSTALLED_MODULES:
        for pkg in {"tasks", "dags", "commands"}:
            pkg_path = ".".join([module, pkg])
            try:
                import_module(pkg_path)
            except ModuleNotFoundError as ex:
                # Only skip error if it is specifically raised due to
                # failing importing the requested pkg path.
                if pkg_path in str(ex.msg):
                    continue
                raise



import gluepy
import setuptools

requirements_base = open("./requirements/base.txt").read().split("\n")
requirements_all = open("./requirements/all.txt").read().split("\n")
requirements_dev = open("./requirements/dev.txt").read().split("\n")
requirements_digitalocean = open("./requirements/digitalocean.txt").read().split("\n")
requirements_gcp = open("./requirements/gcp.txt").read().split("\n")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gluepy",
    version=gluepy.VERSION,
    author="Marcus Lind",
    author_email="marcuslind90@gmail.com",
    description="A framework for data scientists",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gluepy/gluepy",
    packages=setuptools.find_packages(),
    scripts=["gluepy/bin/gluepy-cli.py"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    install_requires=(requirements_base),
    extra_require={
        "all": (requirements_base + requirements_dev + requirements_digitalocean),
        "digitalocean": (requirements_base + requirements_digitalocean),
        "gcp": (requirements_base + requirements_gcp),
    },
    entry_points="""
        [console_scripts]
        gluepy-cli=gluepy.commands.gluepy:cli
    """,
)



import os
import jinja2
import click


@click.group()
def cli():
    """Group of commands for `gluepy` CLI."""


@cli.command()
@click.argument("project")
def startproject(project):
    """Command used to initiate a new Gluepy project.

    Args:
        project (str): project name
    """
    files = {
        f"{project}/manage.py": "manage.j2",
        f"{project}/configs/settings.py": "settings.j2",
        f"{project}/configs/context.yaml": "context.j2",
    }
    for path, template in files.items():
        _create_file_from_jinja(path, template, project=project)

    click.secho(f"Created project '{project}'")


@cli.command()
@click.argument("module")
def startmodule(module):
    """Command used to initiate a new Module in existing Gluepy project.

    Args:
        module (str): module name
    """
    files = {
        f"{module}/dags.py": "dags.j2",
        f"{module}/tasks.py": "tasks.j2",
    }
    for path, template in files.items():
        _create_file_from_jinja(path, template, module=module)

    click.secho(f"Created module '{module}'")


def _create_file_from_jinja(file_path, template, **context):
    env = jinja2.Environment(
        loader=jinja2.PackageLoader("gluepy", "templates"),
        autoescape=jinja2.select_autoescape(),
    )
    t = env.get_template(template)
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, mode="w") as stream:
        stream.write(t.render(**context))



# flake8: noqa
from gluepy.conf import default_settings
from gluepy.utils.loading import LazyProxy, import_string, SingletonMixin
from .local import LocalStorage
from .s3 import S3Storage
from .google import GoogleStorage
from .base import BaseStorage


default_storage: BaseStorage = LazyProxy(
    lambda: import_string(default_settings.STORAGE_BACKEND)()
)
